#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/mount.h>

#define CHUNK_SIZE	4096
#define HOLE_SIZE	128*CHUNK_SIZE

static int
create_sparse_file(const char *fname)
{
	int fd;
	char c = 'C';		/* C like Create */

	if ( (fd = open(fname, O_RDWR | O_TRUNC | O_CREAT, 0666)) < 0) {
		perror("open");
		return -1;
	}

	if (write(fd, &c, 1) != 1) {
		perror("write1");
		goto err_out;
	}

	if (lseek(fd, HOLE_SIZE, SEEK_SET) != HOLE_SIZE) {
		perror("lseek");
		goto err_out;
	}

	if (write(fd, &c, 1) != 1) {
		perror("write2");
		goto err_out;
	}

	return fd;

err_out:
	if (close(fd))
		perror("close");
	return -1;
}

static int
fill_holes(int fd)
{
	off_t pos;
	char *addr;

	for (pos = 0; pos <= HOLE_SIZE; pos += CHUNK_SIZE) {
		addr = mmap(NULL, CHUNK_SIZE, PROT_READ | PROT_WRITE,
			    MAP_SHARED, fd, pos);
		if (addr == MAP_FAILED) {
			perror("mmap");
			return -1;
		}

		*addr = 'F';	/* F like Fill */
		if (munmap(addr, CHUNK_SIZE)) {
			perror("munmap");
			return -1;
		}
	}

	return 0;
}

static int
fill_holes_write(int fd)
{
	off_t pos;
	char c = 'F';		/* F like Fill */

	for (pos = 0; pos <= HOLE_SIZE; pos += CHUNK_SIZE) {
		if (lseek(fd, pos, SEEK_SET) != pos) {
			perror("lseek");
			return -1;
		}

		if (write(fd, &c, 1) != 1) {
			perror("write hole");
			return -1;
		}
	}

	return 0;
}

int main(int argc, char **argv)
{
	int fd;

	if (argc < 3) {
		fprintf(stderr, "Usage: %s <testfile> <reffile>\n", argv[0]);
		return 1;
	}

	if ( (fd = create_sparse_file(argv[1])) < 0)
		return 2;

	if (fill_holes(fd))
		return 2;

	if (close(fd)) {
		perror("close testfile");
		return 2;
	}

	if ( (fd = create_sparse_file(argv[2])) < 0)
		return 2;

	if (fill_holes_write(fd))
		return 2;

	if (close(fd)) {
		perror("close reference file");
		return 2;
	}

	return 0;
}
